package cz.cuni.dsrg.sofa.netbeans.visualdesigner.designer;

import cz.cuni.dsrg.sofa.netbeans.visualdesigner.explorer.nodes.ComponentNode;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects.Architecture;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects.Binding;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects.Component;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects.Interface;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.nodes.Node;

import javax.swing.*;
import java.awt.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;

/**
 * This class represents the centeral functionality of the component
 * designer.<p>
 *
 * It is a JPanel composed of multiple layers, to display the frame,
 * subcomponents, and bindings. The event layer takes care of handling
 * the mouse events.<p>
 *
 * It handles the displayed context and the selection. It implements
 * commands to modify the edited architecture. It is able to layout
 * a commited architecture automatically.
 *
 * @author Michael Cifka
 */
public class ArchitectureEditor extends JPanel implements PropertyChangeListener
{

    public ArchitectureEditor(ArchitectureWindow window)
    {
        this.window = window;

        setLayout(new GridBagLayout());
        setOpaque(true);
        setBackground(new Color(204, 204, 204));

        layeredPane = new JLayeredPane();
        layeredPane.setLayout(new OverlayLayout(layeredPane));

        bindingLayer = new BindingLayer();
        layeredPane.add(bindingLayer, new Integer(20));

        eventLayer = new EventLayer(this);
        layeredPane.add(eventLayer, new Integer(30));

        componentLayer = new JPanel();
        componentLayer.setBackground(Color.white);
        componentLayer.setOpaque(true);
        componentLayer.setLayout(null);

        layeredPane.add(componentLayer, new Integer(10));
        scrollPane = new JScrollPane(layeredPane);

        Actions actions = new Actions(this);
        JToolBar toolBar = actions.createToolBar();

        GridBagConstraints gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.anchor = GridBagConstraints.WEST;
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.insets = new Insets(10, 10, 10, 10);
        add(toolBar, gridBagConstraints);

        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        gridBagConstraints.insets = new Insets(0, 12, 12, 12);
        add(scrollPane, gridBagConstraints);
    }

    public void setContents(ComponentNode node, final Object selObject)
    {
        // remove property change listener from previous contents
        if(this.node != null)
            this.node.getComponent().removePropertyChangeListener(this);
        this.node = node;
        // add property change listener to the current contents
        node.getComponent().addPropertyChangeListener(this);

        SwingUtilities.invokeLater(new Runnable()
        {
            public void run()
            {
                componentLayer.setVisible(false);
                // update the graphical components
                componentLayer.removeAll();
                bindingLayer.removeAll();
                createFrame();
                createArchitecture();
                setSelection(null);
                // restore the selection
                selectObject(selObject);
                eventLayer.setMode(EventLayer.SELECTION_MODE);
                componentLayer.setVisible(true);
            }
        });
    }

    public ArchitectureWindow getArchitectureWindow()
    {
        return window;
    }

    public void propertyChange(PropertyChangeEvent e)
    {
        if(!e.getPropertyName().equals("layout"))
        {
            update();
        }
    }

    public void update()
    {
        // find the selected object
        Object selObject = getObject(selection);
        // rebuild the graphical components
        setContents(this.node, selObject);
    }

    public JComponent getJComponentAt(Point p)
    {
        JComponent c = (JComponent) componentLayer.getComponentAt(p);
        return (c != componentLayer) ? c : null;
    }

    public void setSelection(JComponent c)
    {
        selection = c;

        if(c instanceof GrInstance)
            window.setActivatedNodes(new Node[]{
                node.getChildren().findChild(((GrInstance) c).getInstance().getName())});
        else if(c instanceof GrInterface)
            window.setActivatedNodes(new Node[]{((GrInterface) c).getNode()});
        else
            window.setActivatedNodes(new Node[]{node});
    }

    public JComponent getSelection()
    {
        return selection;
    }

    public GrFrame getFrame()
    {
        return frame;
    }

    public void resetSize(int width, int height)
    {
        Dimension size = new Dimension(width, height);
        componentLayer.setPreferredSize(size);
        eventLayer.setPreferredSize(size);
        layeredPane.revalidate();
    }

    public GrInstance[] getInstances()
    {
        ArrayList list = new ArrayList();
        java.awt.Component[] components = componentLayer.getComponents();
        for(int i = 0; i < components.length; i++)
            if(components[i] instanceof GrInstance)
                list.add(components[i]);
        return (GrInstance[]) list.toArray(new GrInstance[0]);
    }

    public GrInterface[] getInterfaces(JComponent frame)
    {
        ArrayList list = new ArrayList();
        java.awt.Component[] components = componentLayer.getComponents();
        for(int i = 0; i < components.length; i++)
            if(components[i] instanceof GrInterface
                    && (frame == null || ((GrInterface) components[i]).getFrame() == frame))
                list.add(components[i]);
        return (GrInterface[]) list.toArray(new GrInterface[0]);
    }

    // could be implemented more effectively using an index map
    public GrInterface getInterface(Interface iface)
    {
        GrInterface[] grInterfaces = getInterfaces(null);
        for(int i = 0; i < grInterfaces.length; i++)
            if(grInterfaces[i].getInterface() == iface) return grInterfaces[i];
        return null;
    }

    public void createFrame()
    {
        Component component = node.getComponent();

        if(!component.isLaidout())
            layoutFrame(component);

        frame = new GrFrame(component, this);
        Dimension d= componentLayer.getSize();
        frame.setSize(7*d.width/8, 7*d.height/8);
        componentLayer.add(frame);
        Interface[] ifaces = component.getInterfaces();
        for(int i = 0; i < ifaces.length; i++)
            componentLayer.add(new GrInterface(ifaces[i], frame), 0);
    }

    public void createArchitecture()
    {
        Component component = node.getComponent();
        Architecture architecture = component.getImplementation();
        if(architecture == null)
            return; // primitive architecture
        if(!architecture.isLaidout())
        {
            layoutArchitecture(component, architecture);
            frame.setSize(component.getLayout().getStandaloneSize());
        }

        Component[] instances = architecture.getInstances();
        for(int i = 0; i < instances.length; i++)
        {

            GrInstance instFrame = new GrInstance(instances[i], frame, this);
            componentLayer.add(instFrame, 0);
            Interface[] ifaces = instances[i].getInterfaces();
            for(int j = 0; j < ifaces.length; j++)
            {
                if(ifaces[j].getLayout().getEmbeddedLocation().equals(new Point(0, 0)))
                {
                    Point p = findInterfaceLocation(instances[i], false);
                    ifaces[j].getLayout().setEmbeddedLocation(p);
                }
                componentLayer.add(new GrInterface(ifaces[j], instFrame), 0);
            }
        }

        Binding[] bindings = architecture.getBindings();
        for(int i = 0; i < bindings.length; i++)
        {
            GrInterface source = getInterface(bindings[i].getLhsOperand());
            GrInterface target = getInterface(bindings[i].getRhsOperand());
            bindingLayer.addBinding(source, target);
        }

    }

    public void layoutFrame(Component comp)
    {
        comp.getLayout().setStandaloneSize(Constants.FRAME_WIDTH, Constants.FRAME_HEIGHT);
        Interface[] ifaces = comp.getInterfaces();
        for(int i = 0; i < ifaces.length; i++)
        {
            ifaces[i].getLayout().setStandaloneLocation(
                    findInterfaceLocation(comp, true));
        }

        comp.setLaidout(true);
    }

    public void layoutArchitecture(Component component, Architecture architecture)
    {
        Component[] instances = architecture.getInstances();
        if(instances.length == 0) return;

        int rows = (int) Math.ceil(Math.sqrt(instances.length));
        int cols = (int) Math.ceil((double) instances.length / rows);

        int x = Constants.FRAME_INSETS + Constants.INSTANCE_INSETS;
        int y = Constants.FRAME_INSETS + Constants.INSTANCE_INSETS;

        for(int i = 0, j = 0; i < instances.length; i++)
        {
            instances[i].getLayout().setEmbeddedLocation(x, y);
            instances[i].getLayout().setEmbeddedSize(
                    Constants.INSTANCE_WIDTH, Constants.INSTANCE_HEIGHT);

            Interface[] ifaces = instances[i].getInterfaces();
            for(int k = 0; k < ifaces.length; k++)
            {

                ifaces[k].getLayout().setEmbeddedLocation(
                        findInterfaceLocation(instances[i], false));
            }

            if(++j < cols)
            {
                x += Constants.INSTANCE_WIDTH - 1 + Constants.INSTANCE_INSETS;
            }
            else
            {
                j = 0;
                x = Constants.FRAME_INSETS + Constants.INSTANCE_INSETS;
                y += Constants.INSTANCE_HEIGHT - 1 + Constants.INSTANCE_INSETS;
            }

        }
        Dimension frameSize = new Dimension(
                cols * (Constants.INSTANCE_WIDTH - 1)
                + (cols + 1) * Constants.INSTANCE_INSETS,
                rows * (Constants.INSTANCE_HEIGHT - 1)
                + (rows + 1) * Constants.INSTANCE_INSETS);

        Dimension before = component.getLayout().getStandaloneSize();
        component.getLayout().setStandaloneSize(
                Math.max(frameSize.width, before.width),
                Math.max(frameSize.height, before.height));

        architecture.setLaidout(true);
    }

    public void addInstance(Rectangle rect)
    {
        Component comp = node.getComponent();
        Architecture arch = comp.getImplementation();

        // this line might be removed (no effect)
        if(arch == null) arch = comp.createImplementation();

        arch.setLaidout(true);

        if(!checkArchitecture(arch)) return;
        Component inst = arch.createInstance();
        inst.getLayout().setEmbeddedLocation(rect.getLocation());
        inst.getLayout().setEmbeddedSize(rect.getSize());
        arch.addInstance(inst);
        selectObject(inst);
    }

    public static final int INTERFACE_PROVISION = 0;
    public static final int INTERFACE_REQUIREMENT = 1;

    public void setInterfaceMode(int mode)
    {
        interfaceMode = mode;
    }

    public void addInterface(Point location, JComponent frame)
    {
        Interface iface;
        if(frame instanceof GrFrame)
        {
            Component comp = node.getComponent();
            if(!checkFrame(comp)) return;
            iface = comp.createInterface((interfaceMode == INTERFACE_PROVISION) ?
                                         Interface.PROVIDED : Interface.REQUIRED);
            iface.getLayout().setStandaloneLocation(location);
            if(comp.getParent() != null && comp.getParent().isLaidout())
                iface.getLayout().setEmbeddedLocation(
                        findInterfaceLocation(comp, false));
            comp.addInterface(iface);
        }
        else
        {
            Component comp = ((GrInstance) frame).getInstance();
            if(!checkFrame(comp)) return;
            iface = comp.createInterface((interfaceMode == INTERFACE_PROVISION) ?
                                         Interface.PROVIDED : Interface.REQUIRED);
            iface.getLayout().setEmbeddedLocation(location);
            if(comp.isLaidout())
                iface.getLayout().setStandaloneLocation(
                        findInterfaceLocation(comp, true));
            comp.addInterface(iface);
        }
        selectObject(iface);
    }

    public void addBinding(GrInterface ifaceA, GrInterface ifaceB)
    {
        Component comp = node.getComponent();
        Architecture arch = comp.getImplementation();

        if(!checkArchitecture(arch)) return;

        Interface sourceIface, targetIface;
        if(ifaceA.isSource())
        {
            sourceIface = ifaceA.getInterface();
            targetIface = ifaceB.getInterface();
        }
        else
        {
            sourceIface = ifaceB.getInterface();
            targetIface = ifaceA.getInterface();
        }

        if(sourceIface.getType() != targetIface.getType())
        {
            String message = "The interfaces that you are trying to bind"
                    + " have different types.\nWhat do you wish to do?";
            NotifyDescriptor desc = new NotifyDescriptor.Confirmation(message,
                                                                      NotifyDescriptor.DEFAULT_OPTION, NotifyDescriptor.QUESTION_MESSAGE);
            JButton buttonSource = new JButton("Set to " + sourceIface.getType().getName());
            JButton buttonTarget = new JButton("Set to " + targetIface.getType().getName());
            JButton buttonCancel = new JButton("Do not bind");
            desc.setOptions(new Object[]{buttonSource, buttonTarget, buttonCancel});
            Object result = TopManager.getDefault().notify(desc);

            if(result == buttonSource)
            {
                if(!checkFrame(targetIface.getParent())) return;
                targetIface.setType(sourceIface.getType());
            }
            else if(result == buttonTarget)
            {
                if(!checkFrame(sourceIface.getParent())) return;
                sourceIface.setType(targetIface.getType());
            }
            else
                return;
        }

        arch.addBinding(new Binding(sourceIface, targetIface, arch));
    }

    public void delete()
    {
        if(selection instanceof GrInstance)
        {
            Component comp = node.getComponent();
            Architecture arch = comp.getImplementation();
            if(!checkArchitecture(arch)) return;
            GrInstance frame = (GrInstance) selection;
            Component inst = frame.getInstance();
            arch.removeInstance(inst);
        }
        else if(selection instanceof GrInterface)
        {
            GrInterface iface = (GrInterface) selection;
            JComponent frame = iface.getFrame();
            Component comp = (frame instanceof GrInstance)
                    ? ((GrInstance) frame).getInstance() : node.getComponent();
            if(!checkFrame(comp)) return;
            comp.removeInterface(iface.getInterface());
        }
    }

    public void unbind()
    {
        if(selection instanceof GrInterface)
        {
            Interface iface = ((GrInterface) selection).getInterface();
            Architecture arch = node.getComponent().getImplementation();
            if(!checkArchitecture(arch)) return;
            Binding[] binds = iface.getBindings();
            for(int i = 0; i < binds.length; i++)
                arch.removeBinding(binds[i]);
            node.getComponent().firePropertyChange("bindings");
        }
    }


    public EventLayer getEventLayer()
    {
        return eventLayer;
    }

    private Point findInterfaceLocation(Component comp, boolean standalone)
    {
        Dimension size = standalone ? comp.getLayout().getStandaloneSize()
                : comp.getLayout().getEmbeddedSize();
        Point loc = standalone ? new Point(Constants.FRAME_INSETS, Constants.FRAME_INSETS)
                : comp.getLayout().getEmbeddedLocation();
        int rows = size.height / Constants.GRID_SIZE - 1;
        int cols = size.width / Constants.GRID_SIZE - 1;

        // count the intefraces already present in each location
        int[] arr = new int[2 * (rows + cols)];
        Interface[] ifaces = comp.getInterfaces();
        for(int i = 0; i < ifaces.length; i++)
        {
            Point p = standalone ? ifaces[i].getLayout().getStandaloneLocation()
                    : ifaces[i].getLayout().getEmbeddedLocation();
            if(p.y == loc.y)  // NORTH
                arr[(p.x - loc.x) / Constants.GRID_SIZE - 1]++;
            else if(p.x == loc.x) // WEST
                arr[(p.y - loc.y) / Constants.GRID_SIZE - 1 + cols]++;
            else if(p.y == loc.y + size.height - 1) // SOUTH
                arr[(p.x - loc.x) / Constants.GRID_SIZE - 1 + cols + rows]++;
            else if(p.x == loc.x + size.width - 1) // EAST
                arr[(p.y - loc.y) / Constants.GRID_SIZE - 1 + 2 * cols + rows]++;
        }

        int pos = -1;

        // find first even empty position with both surrounding positions empty
        for(int i = 1; i < 2 * (cols + rows); i += 2)
        {
            if(i == cols || i == cols + rows || i == 2 * cols + rows) i++;
            if(arr[i - 1] == 0 && arr[i] == 0
                    && ((i + 1 == cols || i + 1 == cols + rows
                    || i + 1 == 2 * cols + rows || i + 1 == 2 * (cols + rows))
                    || arr[i + 1] == 0))
            {
                pos = i;
                break;
            }
        }

        // if no such position exists, find first position with no or least interfaces
        if(pos == -1)
        {
            int least = 999999;
            for(int i = 0; i < 2 * (cols + rows); i++)
                if(arr[i] < least)
                {
                    pos = i;
                    least = arr[i];
                }
        }

        // compute location
        if(pos < cols) // NORTH
            return new Point(loc.x + (pos + 1) * Constants.GRID_SIZE, loc.y);
        else if(pos < cols + rows) // WEST
            return new Point(loc.x, loc.y + (pos - cols + 1) * Constants.GRID_SIZE);
        else if(pos < 2 * cols + rows) // SOUTH
            return new Point(loc.x + (pos - cols - rows + 1) * Constants.GRID_SIZE, loc.y + size.height - 1);
        else // EAST
            return new Point(loc.x + size.width - 1, loc.y + (pos - 2 * cols - rows + 1) * Constants.GRID_SIZE);
    }

    private Object getObject(JComponent c)
    {
        if(c instanceof GrFrame)
            return node.getComponent();
        else if(c instanceof GrInstance)
            return ((GrInstance) c).getInstance();
        else if(c instanceof GrInterface)
            return ((GrInterface) c).getInterface();
        else
            return null;
    }

    private void selectObject(Object obj)
    {
        if(obj != null)
            for(int i = 0; i < componentLayer.getComponentCount(); i++)
            {
                JComponent current = (JComponent) componentLayer.getComponent(i);
                if(obj == getObject(current))
                {
                    setSelection(current);
                    return;
                }
            }
        else
            setSelection(null);
    }

    public boolean checkArchitecture(Architecture arch)
    {
        if(arch.getType().isCommited())
        {
            NotifyDescriptor d = new NotifyDescriptor.Message(
                    "Cannot modify a commited architecture.",
                    NotifyDescriptor.INFORMATION_MESSAGE);
            TopManager.getDefault().notify(d);
            return false;
        }
        return true;
    }

    public boolean checkFrame(Component comp)
    {
        if(comp.getType().isCommited())
        {
            NotifyDescriptor d = new NotifyDescriptor.Message(
                    "Cannot modify a commited frame.",
                    NotifyDescriptor.INFORMATION_MESSAGE);
            TopManager.getDefault().notify(d);
            return false;
        }
        return true;
    }

    private JLayeredPane layeredPane;
    private EventLayer eventLayer;
    private JComponent selection;
    private JScrollPane scrollPane;
    private JPanel componentLayer;
    private GrFrame frame;
    private BindingLayer bindingLayer;
    private ComponentNode node;
    private ArchitectureWindow window;
    private int interfaceMode;
}