package cz.cuni.dsrg.sofa.netbeans.visualdesigner.designer;

import cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects.Component;

import javax.swing.*;
import java.awt.*;

/**
 * A visually designable component representing a subcomponent.
 *
 * @author Michael Cifka
 */
public class GrInstance extends JPanel implements VisualDesign
{

    public static final int BORDER = Constants.INSTANCE_INSETS;

    public GrInstance(Component inst, GrFrame frame, ArchitectureEditor editor)
    {
        this.inst = inst;
        this.frame = frame;
        this.editor = editor;

        Point location = inst.getLayout().getEmbeddedLocation();
        Dimension size = inst.getLayout().getEmbeddedSize();
        super.setBounds(location.x, location.y, size.width, size.height);

        setOpaque(false);
        setLayout(null);

        labelPanel = new InsidePanel(inst, true);
        add(labelPanel);

        if(inst.getArrayProperty() != null)
        {
            shadowPanel = new InsidePanel(inst, false);
            add(shadowPanel);
        }

        layoutInsidePanels();
    }

    // constructor used for a ghost component
    GrInstance(GrFrame frame)
    {
        this.frame = frame;
        super.setBounds(0, 0, Constants.INSTANCE_WIDTH, Constants.INSTANCE_HEIGHT);
    }

    public Component getInstance()
    {
        return inst;
    }

    public void setBounds(int x, int y, int width, int height)
    {
        Rectangle before = getBounds();
        super.setBounds(x, y, width, height);

        // layout inside panels
        layoutInsidePanels();

        // adjust interfaces
        GrInterface[] ifaces = editor.getInterfaces(this);
        if((x != before.x || y != before.y)
                && width == before.width && height == before.height)
        // if moved, move also interfaces
            for(int i = 0; i < ifaces.length; i++)
            {
                Point location = ifaces[i].getLocation();
                location.translate(x - before.x, y - before.y);
                ifaces[i].setLocation(location);
            }
        else
        // check and correct positions of the interfaces
            for(int i = 0; i < ifaces.length; i++)
                ifaces[i].reposition();


        // update layout info
        inst.getLayout().setEmbeddedLocation(x, y);
        inst.getLayout().setEmbeddedSize(width, height);

        // fire layout event
        inst.getParent().getParent().firePropertyChange("layout");
    }

    public boolean canMove()
    {
        return true;
    }

    public boolean canResize(int direction)
    {
        return true;
    }

    public Rectangle reviewMove(Rectangle before, Rectangle after)
    {
        Rectangle frameBounds = frame.getBounds();
        frameBounds.grow(-BORDER, -BORDER);
        int minx = frameBounds.x, maxx = frameBounds.x + frameBounds.width;
        int miny = frameBounds.y, maxy = frameBounds.y + frameBounds.height;
        if(after.x < minx)
            after.x = minx;
        else if(after.x + after.width > maxx) after.x = maxx - after.width;
        if(after.y < miny)
            after.y = miny;
        else if(after.y + after.height > maxy) after.y = maxy - after.height;
        return after;
    }

    public Rectangle reviewResize(Rectangle before, Rectangle after)
    {
        Rectangle frameBounds = frame.getBounds();
        frameBounds.grow(-BORDER, -BORDER);
        int minx = frameBounds.x, maxx = frameBounds.x + frameBounds.width;
        int miny = frameBounds.y, maxy = frameBounds.y + frameBounds.height;
        if(after.x < minx)
        {
            after.width -= minx - after.x;
            after.x = minx;
        }
        else if(after.x + after.width > maxx) after.width = maxx - after.x;
        if(after.y < miny)
        {
            after.height -= miny - after.y;
            after.y = miny;
        }
        else if(after.y + after.height > maxy) after.height = maxy - after.y;

        // respect interfaces
        GrInterface[] ifaces = editor.getInterfaces(this);
        for(int i = 0; i < ifaces.length; i++)
            after = ifaces[i].reviewFrameResize(after);

        return after;
    }

    public Rectangle reviewFrameResize(Rectangle r)
    {
        Rectangle bounds = new Rectangle(getBounds());
        bounds.grow(BORDER, BORDER);
        return r.union(bounds);
    }

    public Dimension getMinimumSize()
    {
        Dimension d = new Dimension(super.getMinimumSize());
        d.width = (int) Math.ceil(d.width / Constants.GRID_SIZE) * Constants.GRID_SIZE + 1;
        d.height = (int) Math.ceil(d.height / Constants.GRID_SIZE) * Constants.GRID_SIZE + 1;
        return d;
    }

    private void layoutInsidePanels()
    {
        Dimension size = getSize();
        if(shadowPanel == null)
            labelPanel.setBounds(0, 0, size.width, size.height);
        else
        {
            labelPanel.setBounds(0, 0, size.width - 2, size.height - 2);
            shadowPanel.setBounds(2, 2, size.width - 2, size.height - 2);
        }
    }

    private Component inst;
    private GrFrame frame;
    private ArchitectureEditor editor;

    private JPanel labelPanel;
    private JPanel shadowPanel;

    private class InsidePanel extends JPanel
    {
        public InsidePanel(Component inst, boolean label)
        {
            setOpaque(true);
            setBackground(Constants.INSTANCE_COLORS[inst.getColorIndex()]);
            setBorder(BorderFactory.createCompoundBorder(
                    BorderFactory.createLineBorder(Color.black),
                    BorderFactory.createEmptyBorder(8, 8, 8, 8)));
            if(label)
            {
                setLayout(new GridBagLayout());
                JLabel lblName = new JLabel(inst.getName());
                lblName.setFont(new Font("Dialog", 0, 11));
                lblName.setForeground(Color.black);
                GridBagConstraints con = new GridBagConstraints();
                con.gridx = 0;
                con.gridy = 0;
                add(lblName, con);

                JLabel lblFrame = new JLabel("[" + inst.getType().getName() + "]");
                lblFrame.setFont(new Font("Dialog", 0, 11));
                lblFrame.setForeground(Color.black);
                con = new GridBagConstraints();
                con.gridx = 0;
                con.gridy = 1;
                add(lblFrame, con);
            }
        }
    }
}
