package cz.cuni.dsrg.sofa.netbeans.visualdesigner.designer;

import cz.cuni.dsrg.sofa.netbeans.visualdesigner.explorer.nodes.InterfaceNode;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects.Interface;

import javax.swing.*;
import java.awt.*;

/**
 * A visually designable component representing an interface.
 *
 * @author Michael Cifka
 */
public class GrInterface extends JPanel implements VisualDesign
{

    public static final int SIZE = Constants.INTERFACE_SIZE;

    public GrInterface(Interface iface, JComponent frame)
    {
        this.iface = iface;
        this.frame = frame;
        standalone = (frame instanceof GrFrame);
        provision = (iface.getMode() == Interface.PROVIDED);
        node = new InterfaceNode(iface);

        setOpaque(false);
        setLayout(null);
        setToolTipText((iface.getType() == null) ? iface.getName()
                       : iface.getName() + " [" + iface.getType().getName() + "]");

        Point center = standalone ? iface.getLayout().getStandaloneLocation()
                : iface.getLayout().getEmbeddedLocation();
        super.setBounds(center.x - SIZE / 2, center.y - SIZE / 2, SIZE, SIZE);
        updateOrientation();

        InsidePanel interfacePanel = new InsidePanel(provision);
        add(interfacePanel);

        if(iface.getArrayProperty() != null)
        {
            InsidePanel shadowPanel = new InsidePanel(provision);
            add(shadowPanel);
            interfacePanel.setBounds(0, 0, SIZE - 2, SIZE - 2);
            shadowPanel.setBounds(2, 2, SIZE - 2, SIZE - 2);
        }
        else
            interfacePanel.setBounds(0, 0, SIZE, SIZE);

    }

    // constructor used for a ghost component
    GrInterface(int x, int y, int direction)
    {
        super.setBounds(x - SIZE / 2, y - SIZE / 2, SIZE, SIZE);
        this.direction = direction;
    }

    public JComponent getFrame()
    {
        return frame;
    }

    public Interface getInterface()
    {
        return iface;
    }

    public boolean isSource()
    {
        return provision ^ frame instanceof GrInstance;
    }

    public boolean isTarget()
    {
        return !isSource();
    }

    public boolean canBind(GrInterface other)
    {
        return (frame != other.frame) && (isSource() ^ other.isSource());
    }

    public int getDirection()
    {
        return direction;
    }

    public InterfaceNode getNode()
    {
        return node;
    }

    // center location

    public void setCenterLocation(Point p)
    {
        setCenterLocation(p.x, p.y);
    }

    public void setCenterLocation(int x, int y)
    {
        setBounds(x - SIZE / 2, y - SIZE / 2, SIZE, SIZE);
    }

    public Point getCenterLocation()
    {
        return new Point(getX() + SIZE / 2, getY() + SIZE / 2);
    }

    public void setBounds(int x, int y, int width, int height)
    {
        // do not allow resize
        super.setBounds(x, y, SIZE, SIZE);

        // convert to center location
        x += SIZE / 2;
        y += SIZE / 2;

        // update the orientation
        updateOrientation();

        // save the location to the layout info
        if(standalone)
            iface.getLayout().setStandaloneLocation(x, y);
        else
            iface.getLayout().setEmbeddedLocation(x, y);

        // fire layout event
        iface.getParent().firePropertyChange("layout");
    }

    private void updateOrientation()
    {
        Point center = getCenterLocation();
        int minx = frame.getX(), maxx = frame.getX() + frame.getWidth() - 1;
        int miny = frame.getY(), maxy = frame.getY() + frame.getHeight() - 1;
        if(center.x == minx)
            direction = SwingConstants.WEST;
        else if(center.x == maxx)
            direction = SwingConstants.EAST;
        else if(center.y == miny)
            direction = SwingConstants.NORTH;
        else if(center.y == maxy) direction = SwingConstants.SOUTH;
    }

    public void reposition()
    {
        Point c = getCenterLocation();
        int minx = frame.getX(), maxx = frame.getX() + frame.getWidth() - 1;
        int miny = frame.getY(), maxy = frame.getY() + frame.getHeight() - 1;
        if(direction == SwingConstants.WEST)
            c.x = minx;
        else if(direction == SwingConstants.EAST)
            c.x = maxx;
        else if(direction == SwingConstants.NORTH)
            c.y = miny;
        else if(direction == SwingConstants.SOUTH) c.y = maxy;
        setCenterLocation(c);
    }

    public boolean canMove()
    {
        return true;
    }

    public boolean canResize(int direction)
    {
        return false;
    }

    public Rectangle reviewMove(Rectangle before, Rectangle after)
    {
        // get central point of interface bounds
        Point centerBefore = before.getLocation();
        centerBefore.translate(SIZE / 2, SIZE / 2);
        Point centerAfter = after.getLocation();
        centerAfter.translate(SIZE / 2, SIZE / 2);

        // get motion difference and frame bounds
        int dx = centerAfter.x - centerBefore.x;
        int dy = centerAfter.y - centerBefore.y;
        int minx = frame.getX(), maxx = frame.getX() + frame.getWidth() - 1;
        int miny = frame.getY(), maxy = frame.getY() + frame.getHeight() - 1;

        // allow motion only along edges
        if(centerBefore.x != minx && centerBefore.x != maxx) dy = 0;
        if(centerBefore.y != miny && centerBefore.y != maxy) dx = 0;
        Point centerResult = new Point(centerBefore.x + dx, centerBefore.y + dy);

        // do not allow motion outsize frame bounds
        if(centerResult.x < minx)
            centerResult.x = minx;
        else if(centerResult.x > maxx) centerResult.x = maxx;
        if(centerResult.y < miny)
            centerResult.y = miny;
        else if(centerResult.y > maxy) centerResult.y = maxy;

        // do not allow corner positons
        if((centerResult.x == minx || centerResult.x == maxx)
                && (centerResult.y == miny || centerResult.y == maxy))
        {
            dx = centerAfter.x - centerResult.x;
            dy = centerAfter.y - centerResult.y;
            if(centerResult.x == minx) dx = -dx;
            if(centerResult.y == miny) dy = -dy;
            if(dx > dy)
            {
                if(centerResult.y == miny)
                    centerResult.y += Constants.GRID_SIZE;
                else
                    centerResult.y -= 10;
            }
            else
            {
                if(centerResult.x == minx)
                    centerResult.x += Constants.GRID_SIZE;
                else
                    centerResult.x -= 10;
            }
        }

        // return new interface bounds
        centerResult.translate(-SIZE / 2, -SIZE / 2);
        return new Rectangle(centerResult.x, centerResult.y, SIZE, SIZE);
    }

    public Rectangle reviewResize(Rectangle before, Rectangle after)
    {
        return null;
    }

    public Rectangle reviewFrameResize(Rectangle r)
    {
        Point center = getCenterLocation();
        if(direction == SwingConstants.WEST
                || direction == SwingConstants.EAST)
            return r.union(new Rectangle(
                    r.x, center.y - Constants.GRID_SIZE, 1, 2 * Constants.GRID_SIZE));
        else
            return r.union(new Rectangle(
                    center.x - Constants.GRID_SIZE, r.y, 2 * Constants.GRID_SIZE, 1));
    }

    private Interface iface;
    private JComponent frame;
    private int direction;
    private boolean standalone;
    private boolean provision;
    private InterfaceNode node;

    private class InsidePanel extends JLabel
    {
        public InsidePanel(boolean provision)
        {
            setOpaque(true);
            setBackground(provision ? Color.white : Color.cyan);
            setBorder(BorderFactory.createLineBorder(Color.black));
        }
    }
}
