package cz.cuni.dsrg.sofa.netbeans.visualdesigner.explorer.nodes;


import cz.cuni.dsrg.sofa.netbeans.visualdesigner.designer.ArchitectureWindow;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.explorer.actions.ActGenerateCDL;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.explorer.actions.ActResetComponent;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.explorer.cookies.ComponentCookie;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.explorer.cookies.DataObjectCookie;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.explorer.properties.*;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.loaders.ComponentDataObject;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects.Architecture;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects.Component;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.typemanager.TypeManager;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.actions.DeleteAction;
import org.openide.actions.OpenAction;
import org.openide.actions.PropertiesAction;
import org.openide.actions.RenameAction;
import org.openide.cookies.OpenCookie;
import org.openide.loaders.DataObject;
import org.openide.nodes.*;
import org.openide.util.WeakListener;
import org.openide.util.actions.SystemAction;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * This node represents a component (or a subcomponent). A large set
 * of operations and properties to manipulate the represented component
 * is provided.
 *
 * @author Michael Cifka
 */
public class ComponentNode extends AbstractNode
        implements ComponentCookie, DataObjectCookie, OpenCookie
{

    Component component;
    ComponentDataObject obj;
    PropertyChangeListener listener;
    TypeManager typeManager;

    public ComponentNode(Component component, ComponentDataObject obj)
    {
        super(new Children.Array());
        this.component = component;
        this.obj = obj;
        typeManager = TypeManager.get(component.getRepository());
        setIconBase(component.getParent() == null
                    ? "cz/cuni/dsrg/sofa/netbeans/resources/visualdesigner/icons/component"
                    : "cz/cuni/dsrg/sofa/netbeans/resources/visualdesigner/icons/subcomponent");
        setDefaultAction(SystemAction.get(OpenAction.class));
        super.setName(component.getName());
        updateDisplayName();
        updateChildren();
        getCookieSet().add(this);
        listener = createListener();
        component.addPropertyChangeListener(
                WeakListener.propertyChange(listener, component));
    }

    public Component getComponent()
    {
        return component;
    }

    public DataObject getDataObject()
    {
        return obj;
    }

    public boolean canRename()
    {
        return true;
    }

    public boolean canDestroy()
    {
        Architecture parent = component.getParent();
        return (parent != null || obj.isDeleteAllowed());
    }

    public void setName(String s)
    {
        if(getName().equals(s)) return;

        Architecture parent = component.getParent();
        if(parent != null && parent.getType().isCommited())
        {
            NotifyDescriptor d = new NotifyDescriptor.Message(
                    "Cannot modify a commited architecture.",
                    NotifyDescriptor.INFORMATION_MESSAGE);
            TopManager.getDefault().notify(d);
            return;
        }

        boolean valid = false;

        if(parent == null)
        {
            // try to rename file
            try
            {
                obj.rename2(s);
                valid = true;
            }
            catch (java.io.IOException e)
            {
                valid = false;
            }
        }
        else
        {
            // check if the component can be renamed
            valid = (parent.getInstance(s) == null);
        }

        if(valid)
            component.setName(s);
        else
        {
            NotifyDescriptor d = new NotifyDescriptor.Message(
                    "Cannot rename component " + getName() + " to " + s + ".",
                    NotifyDescriptor.INFORMATION_MESSAGE);
            TopManager.getDefault().notify(d);
            return;
        }

        super.setName(s);
        updateDisplayName();
        firePropertyChange("name", null, null);
    }

    public void destroy() throws java.io.IOException
    {

        Architecture parent = component.getParent();
        ComponentNode parentNode = (parent == null)
                ? null : (ComponentNode) getParentNode();
        ArchitectureWindow editor = ArchitectureWindow.getDefault();

        // find out wheather the node or its subnode
        // is displayed in the architecture editor
        ComponentNode cn = editor.getEditedComponent();
        while(cn != this && cn != null)
        {
            Node n = cn.getParentNode();
            cn = (n instanceof ComponentNode) ? (ComponentNode) n : null;
        }
        // if cn == this, then yes; if cn == null then no
        boolean displayed = (cn == this);

        if(parent == null)
        {
            if(obj.isDeleteAllowed())
                obj.delete2();
        }
        else
        {
            parent.removeInstance(component);
        }
        if(displayed)
        {
            if(parent == null)
                editor.close();
            else
                editor.setEditedComponent(parentNode);
        }
        super.destroy();
    }

    public void updateDisplayName()
    {
        String displayName = component.getName();
        if(component.getType() == null)
            displayName = displayName + " [(not specified)]";
        else
            displayName = displayName + " [" + component.getType().getName() + "]";
        setDisplayName(displayName);
    }

    public void updateChildren()
    {

        Children children = getChildren();
        Node[] nodes = children.getNodes();


        Architecture arch = component.getImplementation();
        if(arch == null)
        {
            children.remove(nodes);
            return;
        }

        Component[] subs = arch.getInstances();

        // remove nodes:
        // get list of subcomponents
        // go through all nodes, check if their subcomponent is in the list
        // if not, then remove the node

        for(int i = 0, j; i < nodes.length; i++)
        {
            Component nc = ((ComponentNode) nodes[i]).getComponent();
            for(j = 0; j < subs.length; j++)
                if(nc == subs[j]) break;
            if(j == subs.length)   // surplus node -- remove
                children.remove(new Node[]{nodes[i]});
        }

        // add nodes
        // go through subcomponentns, chcesk if there is a node for each of
        // them, if isn't then create one and add

        for(int i = 0, j; i < subs.length; i++)
        {
            for(j = 0; j < nodes.length; j++)
            {
                Component nc = ((ComponentNode) nodes[j]).getComponent();
                if(subs[i] == nc) break;
            }
            if(j == nodes.length) // node does not exist -- add
                children.add(new Node[]{new ComponentNode(subs[i], obj)});
        }
    }

    protected SystemAction[] createActions()
    {
        if(getParentNode() == null)
            return new SystemAction[]{
                SystemAction.get(OpenAction.class),
                null,
                SystemAction.get(ActGenerateCDL.class),
                null,
                SystemAction.get(ActResetComponent.class),
                null,
                SystemAction.get(DeleteAction.class),
                SystemAction.get(RenameAction.class),
                null,
                SystemAction.get(PropertiesAction.class)
            };
        else
            return new SystemAction[]{
                SystemAction.get(OpenAction.class),
                null,
                SystemAction.get(DeleteAction.class),
                SystemAction.get(RenameAction.class),
                null,
                SystemAction.get(PropertiesAction.class)
            };
    }


    protected Sheet createSheet()
    {
        //Sheet s = Sheet.createDefault();
        //Sheet.Set set = s.get(Sheet.PROPERTIES);

        Sheet s = new Sheet();

        Sheet.Set setGeneral = createSet(s, "general", "General");
        Sheet.Set setFrame = createSet(s, "frame", "Frame");
        Sheet.Set setArchitecture = createSet(s, "architecture", "Architecture");
        Sheet.Set setValues = createSet(s, "values", "Values");
        Sheet.Set setProject = createSet(s, "project", "Project");


        setGeneral.put(new NameProp());
        setGeneral.put(new PropLocation(component));
        setGeneral.put(new PropSourceFile(component));

        setFrame.put(new PropFrameType(component));
        setFrame.put(new PropFrameArray(component));
        if(component.getType() != null)
        {
            setFrame.put(new PropProperties(component, "properties", "Properties"));
            // add custom frame properties
            String[] properties = component.getProperties();
            for(int i = 0; i < properties.length; i++)
            {
                String type = component.getPropertyType(properties[i]);
                if(type.equals("string"))
                    setValues.put(new PropStringValue(component, properties[i]));
                else if(type.equals("integer"))
                    setValues.put(new PropIntegerValue(component, properties[i]));
            }
        }
        setFrame.put(new PropFrameProtocol(component));

        setArchitecture.put(new PropArchitecture(component));
        Architecture architecture = component.getImplementation();
        if(architecture != null && architecture.getType() != null)
        {
            setArchitecture.put(new PropProperties(architecture, "properties", "Properties"));
            // add custom architecture properties
            String[] properties = architecture.getProperties();
            for(int i = 0; i < properties.length; i++)
            {
                String type = architecture.getPropertyType(properties[i]);
                if(type.equals("string"))
                    setValues.put(new PropStringValue(architecture, properties[i]));
                else if(type.equals("integer"))
                    setValues.put(new PropIntegerValue(architecture, properties[i]));
            }
        }

        setProject.put(new PropCustomCode(component.getRepository()));
        setProject.put(new PropDefaultModule(component.getRepository()));
        setProject.put(new PropDefaultProvider(component.getRepository()));

        return s;
    }

    private Sheet.Set createSet(Sheet s, String name, String displayName)
    {
        Sheet.Set set = new Sheet.Set();
        set.setName(name);
        set.setDisplayName(displayName);
        s.put(set);
        return set;
    }

    private void refreshSheet()
    {
        Sheet s = createSheet();
        setSheet(s);
    }

    public void open()
    {
        ArchitectureWindow editor = ArchitectureWindow.getDefault();
        editor.setEditedComponent(this);
        editor.open();
        editor.setEditedComponent(this);
        editor.requestFocus();

    }

    public void updateEditor()
    {
        ArchitectureWindow editor = ArchitectureWindow.getDefault();
        if(editor.isOpened())
        {
            editor.setEditedComponent(this);
        }
    }

    private PropertyChangeListener createListener()
    {
        return new PropertyChangeListener()
        {
            public void propertyChange(PropertyChangeEvent propertyChangeEvent)
            {
                getDataObject().setModified(true);
                if(!propertyChangeEvent.getPropertyName().equals("layout"))
                {
                    firePropertyChange("frame", null, null);
                    firePropertyChange("architecture", null, null);
                    updateDisplayName();
                    updateChildren();
                    refreshSheet();
                    typeManager.refresh();
                }
            }
        };
    }

    private final class NameProp extends PropertySupport.ReadWrite
    {
        public NameProp()
        {
            super("name", String.class, "Name", null);
        }

        public Object getValue()
        {
            return component.getName();
        }

        public void setValue(Object val)
        {
            ComponentNode.this.setName((String) val);
        }
    }
}


