package cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects;

import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.ArchitectureType;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.FrameType;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.Property;

import java.util.*;

/**
 * This class represents an architecture, the internal structure of a
 * software component.<p>
 *
 * Method to set the type and implementation is provided as well as methods
 * to manage the sets of subcomponent, bindngs, and properties.<p>
 *
 * The architecture is able to build up itself based on a provided type.<p>
 *
 * This class can be persisted using the Java Object Serialization.
 *
 * @author Michael Cifka
 */
public class Architecture implements java.io.Serializable, PropertiesProvider
{

    public Architecture()
    {
        type = null;
        parent = null;
        instances = new LinkedList();
        bindings = new LinkedList();
        instancesIndex = new HashMap();
        propertyTypes = new TreeMap();
        propertyValues = new HashMap();
        laidout = false;
    }

    public Architecture(ArchitectureType type, Component parent)
    {
        this();
        this.type = type;
        this.parent = parent;
        ArchitectureType.Instance[] insts = type.getInstances();
        for(int i = 0; i < insts.length; i++)
        {
            Component inst = new Component(insts[i].getName(), insts[i].getType(), this);
            if(insts[i].getArray() != null)
                inst.setArrayProperty(insts[i].getArray().getProperty());
            instancesIndex.put(inst.getName(), inst);
            instances.add(inst);

        }
        ArchitectureType.Binding[] binds = type.getBindings();
        for(int i = 0; i < binds.length; i++)
        {
            // find lhs interface object
            Component lhsComponent = (binds[i].getMode() == ArchitectureType.Binding.DELEGATE) ?
                    parent : getInstance(binds[i].getLhsInstance());
            Interface lhsInterface = lhsComponent.getInterface(binds[i].getLhsInterface());
            // find rhs interface object
            Component rhsComponent = binds[i].getMode() == ArchitectureType.Binding.SUBSUME ?
                    parent : getInstance(binds[i].getRhsInstance());
            Interface rhsInterface = rhsComponent.getInterface(binds[i].getRhsInterface());
            bindings.add(new Binding(binds[i].getMode(), lhsInterface, rhsInterface, this));
        }
        Property[] properties = type.getProperties();
        for(int i = 0; i < properties.length; i++)
        {
            addProperty(properties[i].getName(), properties[i].getType());
        }
        type.addReference();
    }

    public ArchitectureType getType()
    {
        return type;
    }

    public Component getParent()
    {
        return parent;
    }

    public Component[] getInstances()
    {
        return (Component[]) instances.toArray(new Component[0]);
    }

    public Component getInstance(String name)
    {
        return (Component) instancesIndex.get(name);
    }

    public void addInstance(Component inst)
    {
        // add instance to the collection
        instancesIndex.put(inst.getName(), inst);
        instances.add(inst);
        // modify the associated type
        type.addInstance(inst.getName(), inst.getType());
        getParent().firePropertyChange("instances");
    }

    public Component createInstance()
    {
        FrameType instType = getParent().getRepository().createFrameType();

        String base = "inst";
        int i = 1;
        String s = base + i++;
        while(getInstance(s) != null)
        {
            s = base + i++;
        }
        Component inst = new Component(s, instType, this);
        return inst;
    }

    public void removeInstance(Component inst)
    {
// remove the bindings to this instance
        Interface[] ifaces = inst.getInterfaces();
        for(int i = 0; i < ifaces.length; i++)
        {
            Binding[] binds = ifaces[i].getBindings();
            for(int j = 0; j < binds.length; j++)
                removeBinding(binds[j]);
        }

// modify references
        inst.removeReferences();

        // remove instance from the collection
        instancesIndex.remove(inst.getName());
        instances.remove(inst);

        // modify the associated type
        type.removeInstance(inst.getName());
        getParent().firePropertyChange("instances");
    }

    void renameInstance(Component inst, String name)
    {
        String old = inst.getName();
        instancesIndex.remove(old);
        instancesIndex.put(name, inst);
        inst.rename(name);

        //modify the associated type
        type.renameInstance(old, name);

        getParent().firePropertyChange("instances");
    }

    public Binding[] getBindings()
    {
        return (Binding[]) bindings.toArray(new Binding[0]);
    }

    public void addBinding(Binding bind)
    {
        // add binding to the collection
        bindings.add(bind);
        // modify the associated type
        String lhsInstanceName = bind.getMode() == Binding.DELEGATE ?
                null : bind.getLhsOperand().getParent().getName();
        String rhsInstanceName = bind.getMode() == Binding.SUBSUME ?
                null : bind.getRhsOperand().getParent().getName();
        type.addBinding(lhsInstanceName, bind.getLhsOperand().getName(),
                        rhsInstanceName, bind.getRhsOperand().getName());
        getParent().firePropertyChange("bindings");
    }

    public void removeBinding(Binding bind)
    {
        // remove binding from the collection
        bindings.remove(bind);

// remove binding from the interface
        bind.getLhsOperand().removeBinding(bind);
        bind.getRhsOperand().removeBinding(bind);

        // modify the associated type
        String lhsInstanceName = bind.getMode() == Binding.DELEGATE ?
                null : bind.getLhsOperand().getParent().getName();
        String rhsInstanceName = bind.getMode() == Binding.SUBSUME ?
                null : bind.getRhsOperand().getParent().getName();
        type.removeBinding(lhsInstanceName, bind.getLhsOperand().getName(),
                           rhsInstanceName, bind.getRhsOperand().getName());
    }

    public boolean isLaidout()
    {
        return laidout;
    }

    public void setLaidout(boolean b)
    {
        laidout = b;
    }

    private ArchitectureType type;
    private Component parent;
    private Collection instances;
    private Collection bindings;
    private Map instancesIndex;
    private boolean laidout;

    // PROPERTIES

    private Map propertyTypes;
    private Map propertyValues;

    public String[] getProperties()
    {
        return (String[]) propertyTypes.keySet().toArray(new String[0]);
    }

    public void addProperty(String name, String type)
    {
        propertyTypes.put(name, type);
        propertyValues.remove(name);
        getType().addProperty(name, type);
        getParent().firePropertyChange("properties");
    }

    public void removeProperty(String name)
    {
        propertyTypes.remove(name);
        propertyValues.remove(name);
        getType().removeProperty(name);
        getParent().firePropertyChange("properties");
    }

    public String getPropertyType(String name)
    {
        return (String) propertyTypes.get(name);
    }

    public void setPropertyType(String name, String type)
    {
        addProperty(name, type);
    }

    public Object getPropertyValue(String name)
    {
        Object value = propertyValues.get(name);
        if(value == null)
        {
            // DEFAULT VALUES
            String type = getPropertyType(name);
            if(type == null)
                return null;
            if(type.equals("string"))
                setPropertyValue(name, "");
            else if(type.equals("integer"))
                setPropertyValue(name, new Integer(0));
            else
                return null;
            value = getPropertyValue(name);
        }
        return value;
    }

    public String getPropertyValueAsString(String name)
    {
        Object value = getPropertyValue(name);
        if(value instanceof String) return (String) value;
        if(value instanceof Integer) return "" + ((Integer) value).intValue();
        return "";
    }

    public void setPropertyValue(String name, Object value)
    {
        propertyValues.put(name, value);
    }


    // REFERENCES
    public void removeReferences()
    {
        Component[] insts = getInstances();
        for(int i = 0; i < insts.length; i++)
            insts[i].removeReferences();

        if(type != null)
            type.removeReference();
    }
}