package cz.cuni.dsrg.sofa.netbeans.visualdesigner.types;

import SOFA.SOFAnode.Made.TIR.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.TreeMap;

/**
 * This class is an abstract class from which the classes holding
 * the type information are derived.<p>
 *
 * It provides the functionallity to hold and manage a set of
 * children, and perform various lookups based on the name and
 * version.<p>
 *
 * To automatically remove the type information that is neither
 * used, nor commited, a reference counting algorithm is implemented.<p>
 *
 * Importing the commited contents of the type repository, as well
 * as exporting the uncommited contents as a Component Description
 * Language (.cdl) file is supported.<p>
 *
 * Descedants of this class can be persisted using the Java Object Serialization.
 *
 * @author Michael Cifka
 */
public abstract class AbstractType implements java.io.Serializable
{


    static final long serialVersionUID = -6223471580879995346L;

    AbstractType()
    {
        name = "";
        version = null;
        commited = false;
        container = true;
        parent = null;
        root = null;
    }


    // NAMES

    public String getName()
    {
        return this.name;
    }

    public String getVersion()
    {
        return version;
    }

    public String getVersionString()
    {
        return (version == null) ? "" : "rev \"" + version + "\"";
    }

    public String getScope()
    {
        return parent.getScopeName() + "::";
    }

    public String getNameVersion()
    {
        return name + " " + getVersionString();
    }

    public String getScopeName()
    {
        return getScope() + name;
    }

    public String getScopeNameVersion()
    {
        return getScope() + name + " " + getVersionString();
    }

    public boolean setName(String newName)
    {
        if(name == newName)
            return true;
        if(parent != null)
        {
            if(parent.getChild(newName, null) != null)
                return false;
            AbstractType p = parent;
            p.removeChild(this);
            name = newName;
            p.addChild(this);
        }
        else
            name = newName;
        return true;
    }

    // DEPRECIATED FUNCTIONS

    public String getFullName()
    {
        if(parent instanceof Repository)
            return "::" + getName() + " ver. " + getVersion();
        else
            return parent.getFullName() + "::" + getName() + " ver. " + getVersion();
    }

    public String getRelativeName(AbstractType module)
    {
        String relativeName;
        if(module == parent)
            relativeName = getName();
        else if(parent instanceof Repository || parent instanceof Provider)
            relativeName = "::" + getName();
        else
            relativeName = parent.getRelativeName(null) + "::" + getName();
        if(!isDefaultVersion())
            relativeName += " " + getVersionString();
        return relativeName;
    }


    // REFERENCES

    public void addReference()
    {
        refCounter++;
    }

    public void removeReference()
    {
        refCounter--;
        if(refCounter == 0 && !container && !commited && parent != null)
            parent.removeChild(this);
    }

    public boolean isRemoveable()
    {
        if(refCounter != 0 || commited) return false;
        if(childrenMap.values().size() == 0) return true;
        AbstractType[] ch = getChildren();
        for(int i = 0; i < ch.length; i++)
            if(!ch[i].isRemoveable()) return false;
        return true;
    }


    // OTHER FUNCTIONS

    public boolean isCommited()
    {
        return commited;
    }

    public boolean isContainer()
    {
        return container;
    }

    public AbstractType getParent()
    {
        return parent;
    }

    public Repository getRoot()
    {
        return root;
    }


    // CHILDREN DATA

    private TreeMap childrenMap = new TreeMap();
    private HashMap childrenVersions = new HashMap();

    public boolean addChild(AbstractType child)
    {
        String strName = child.getName();
        String strVersion = child.getVersionString();
        String strNameVersion = child.getNameVersion();
        if(childrenMap.get(strNameVersion) != null) return false;
        childrenMap.put(strNameVersion, child);
        TreeMap versionsForName = (TreeMap) childrenVersions.get(strName);
        if(versionsForName == null)
        {
            versionsForName = new TreeMap();
            childrenVersions.put(strName, versionsForName);
        }
        // uncommited versions are stored under "~" key so that they would
        // be sorted by the tree map to the last place
        if(strVersion.equals("")) strVersion = "~";
        versionsForName.put(strVersion, child);

        // perform additional initializations
        child.parent = this;
        child.root= root;


        return true;
    }

    public void removeChild(AbstractType child)
    {
        String strName = child.getName();
        String strVersion = child.getVersionString();
        String strNameVersion = child.getNameVersion();
        childrenMap.remove(strNameVersion);
        TreeMap versionsForName = (TreeMap) childrenVersions.get(strName);
        if(versionsForName != null)
        {
            if(strVersion.equals("")) strVersion = "~";
            versionsForName.remove(strVersion);
            if(versionsForName.size() == 0) childrenVersions.remove(strName);
        }

// remove the references
        child.parent = null;
        child.root = null;
    }

    public void removeAllChildren()
    {
        childrenMap.clear();
        childrenVersions.clear();
    }

    public AbstractType getChild(String strName, String strVersion)
    {
        TreeMap versionsForName = (TreeMap) childrenVersions.get(strName);
        if(versionsForName == null) return null;
        // uncommited versions are stored under "~" key
        if(strVersion == null)
            return (AbstractType) versionsForName.get("~");
        else
            return (AbstractType) versionsForName.get(strVersion);
    }

    public AbstractType[] getChildren()
    {
        return (AbstractType[]) childrenMap.values().toArray(new AbstractType[0]);
    }

    public AbstractType[] getChildrenForName(String strName)
    {
        TreeMap versionsForName = (TreeMap) childrenVersions.get(strName);
        if(versionsForName == null) return new AbstractType[0];
        return (AbstractType[]) versionsForName.values().toArray(new AbstractType[0]);
    }

    public AbstractType[] getDefaultChildren()
    {
        ArrayList lstTypes = new ArrayList();
        String[] arrNames = (String[]) childrenVersions.keySet().toArray(new String[0]);
        for(int i = 0; i < arrNames.length; i++)
            lstTypes.add(getDefaultChild(arrNames[i]));
        return (AbstractType[]) lstTypes.toArray(new AbstractType[0]);
    }

    public AbstractType getDefaultChild(String strName)
    {
        TreeMap versionsForName = (TreeMap) childrenVersions.get(strName);
        if(versionsForName == null) return null;

        // uncommited versions are now stored under "~" key

        String strDefaultKey = (String) versionsForName.lastKey();
        return (AbstractType) versionsForName.get(strDefaultKey);

    }

    public boolean isDefaultVersion()
    {
        return parent.getDefaultChild(name) == this;
    }

    // DUPLICATION

    public AbstractType duplicate(AbstractType parent)
    {
        return null;
    }

    String getAvailableName(String base, Class clazz)
    {
        String s = base;
        int i = 1;
        AbstractType t = getDefaultChild(s);
        while(t != null && (!t.isCommited() || t.isContainer()
                || !clazz.isInstance(t)))
        {
            s = base + "_" + i++;
            t = getDefaultChild(s);
        }
        return s;
    }

    public String writeAsCDL()
    {
        return writeAsCDL(0);
    }

    String writeAsCDL(int indent)
    {
        return "";
    }

    String indentString(int indent)
    {
        String s = "";
        for(int i = 0; i < indent; i++)
            s = s + "   ";
        return s;
    }

    String indentCode(String code, int indent)
    {
        String indentString = indentString(indent), s = "";
        int start = 0;
        while(true)
        {
            int end = code.indexOf("\n", start);
            if(end == -1) break;
            s += indentString + code.substring(start, end) + "\n";
            start = end + 1;
        }
        if(start < code.length())
            s += indentString + code.substring(start) + "\n";
        return s;
    }

    void destroy()
    {
        if(!commited && parent != null)
        {
            parent.removeChild(this);
        }
    }

    protected String name;
    protected String version;
    protected boolean commited;
    protected boolean container;
    protected AbstractType parent;
    protected Repository root;
    protected int refCounter = 0;


    protected void loadCommitedObjects(TIRObject o)
    {
        try
        {
            Container cont = (Container) o;
            Contained objs[] = cont.contents(null);
            for(int i = 0; i < objs.length; i++)
            {
// load object id and version
                Identification id = objs[i].get_identification();
                String idName = id.name();
                String idVersion = id.version();
                if(idVersion.equals("")) idVersion = null;

// already in the project?
                AbstractType curObj = getChild(idName, idVersion);
                if(curObj == null)
                {
// existing in the project and just commited?
                    curObj = getChild(idName, "null");
                    if(curObj != null)
                    {
                        curObj.commited = true;
                        curObj.version = idVersion;
                    }
                }
                ;
                if(curObj == null)
                {
// create a new type
                    curObj = createAbstractType(objs[i]);
                    if(curObj != null)
                    {
                        curObj.name = idName;
                        curObj.version = idVersion;
                        curObj.commited = true;
                        curObj.parent = this;
                        curObj.root = root;
                        root.indexObject(curObj.getFullName(), curObj);
                        addChild(curObj);
                    }
                }
                if(curObj != null)
                {
// load the children recursively
                    try
                    {
                        curObj.loadCommitedObjects(objs[i]);
                    }
                    catch (NotSupportedException ex)
                    {
                    }
                    ;
                }
            }
        }
        catch (Exception e)
        {
            System.out.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    private AbstractType createAbstractType(Contained obj)
    {
        if(obj instanceof ModuleDef) return new Module();
        if(obj instanceof ProviderDef) return new Provider();
        if(obj instanceof AModuleDef) return new AModule();
        if(obj instanceof InterfaceDef) return new InterfaceType();
        if(obj instanceof FrameDef) return new FrameType();
        if(obj instanceof ArchitectureDef) return new ArchitectureType();
        if(obj instanceof TypedefDef) return new ArrayType();
        return null;
    }

    protected class NotSupportedException extends RuntimeException
    {
    };
}