package cz.cuni.dsrg.sofa.netbeans.visualdesigner.types;

import SOFA.SOFAnode.Made.TIR.*;

import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

/**
 * This class extends the {@link AbstractType} to hold, import, and export
 * the type information representing a frame (i.e. a component type).<p>
 *
 * Reference to the required and provided interfaces are managed encapsulated
 * using the {@link FrameType.Interface} class.<p>
 *
 * This class can be persisted using the Java Object Serialization.
 *
 * @author Michael Cifka
 */
public class FrameType extends AbstractType implements java.io.Serializable
{

    static final long serialVersionUID = -7844112289282511699L;

    public static final int PROVISION = 0;
    public static final int REQUIREMENT = 1;

    public class Interface implements java.io.Serializable
    {
        public Interface(String name, String typeFullName, String arrayFullName, int mode)
        {
            this.name = name;
            this.typeFullName = typeFullName;
            this.arrayFullName = arrayFullName;
            this.mode = mode;
        }

        public Interface(String name, InterfaceType type, int mode)
        {
            this.name = name;
            this.type = type;
            this.mode = mode;
        }

        public String getName()
        {
            return name;
        }

        public InterfaceType getType()
        {
            if(type == null)
                type = (InterfaceType) root.findObject(typeFullName);
            return type;
        }

        public void setType(InterfaceType type)
        {
            this.type = type;
        }

        public ArrayType getArray()
        {
            if(array == null)
                array = (ArrayType) root.findObject(arrayFullName);
            return array;
        }

        public void setArray(ArrayType array)
        {
            arrayFullName = null;
            this.array = array;
        }

        public int getMode()
        {
            return mode;
        }

        public void setMode(int mode)
        {
            this.mode = mode;
        }

        void setName(String name)
        {
            this.name = name;
        }

        private String name;
        private InterfaceType type;
        private ArrayType array;
        private String typeFullName;
        private String arrayFullName;
        private int mode;
    }

    FrameType()
    {
        container = false;
        interfaces = new LinkedList();
        interfacesIndex = new HashMap();
        properties = new HashMap();
        arrayTypes = new HashMap();
    }

    public FrameType(String name)
    {
        this();
        this.name = name;
    }

    private FrameType(String name, FrameType other)
    {
        this(name);
        Interface[] ifaces = other.getInterfaces();
        for(int i = 0; i < ifaces.length; i++)
            addInterface(ifaces[i].getName(),
                         ifaces[i].getType(), ifaces[i].getMode());
    }

    public AbstractType duplicate(AbstractType parent)
    {
        return new FrameType(parent.getAvailableName(name, FrameType.class), this);
    }


    public Interface[] getInterfaces()
    {
        return (Interface[]) interfaces.toArray(new Interface[0]);
    }

    public Interface[] getProvisions()
    {
        Interface[] ifaces = getInterfaces();
        Collection provisions = new LinkedList();
        for(int i = 0; i < ifaces.length; i++)
            if(ifaces[i].getMode() == PROVISION)
                provisions.add(ifaces[i]);
        return (Interface[]) provisions.toArray(new Interface[0]);
    }

    public Interface[] getRequirements()
    {
        Interface[] ifaces = getInterfaces();
        Collection requirements = new LinkedList();
        for(int i = 0; i < ifaces.length; i++)
            if(ifaces[i].getMode() == REQUIREMENT)
                requirements.add(ifaces[i]);
        return (Interface[]) requirements.toArray(new Interface[0]);
    }

    public Interface getInterface(String name)
    {
        return (Interface) interfacesIndex.get(name);
    }

    public void addInterface(String name, InterfaceType type, int mode)
    {
        Interface iface = new Interface(name, type, mode);
        interfacesIndex.put(name, iface);
        interfaces.add(iface);
    }

    public void removeInterface(String name)
    {
        Object obj = interfacesIndex.remove(name);
        interfaces.remove(obj);
    }


    public void renameInterface(String oldName, String newName)
    {
        Interface iface = (Interface) interfacesIndex.remove(oldName);
        if(iface != null)
        {
            interfacesIndex.put(newName, iface);
            iface.setName(newName);
        }
    }

    public Property[] getProperties()
    {
        return (Property[]) properties.values().toArray(new Property[0]);
    }

    public Property getProperty()
    {
        return (Property) properties.get(name);
    }

    public void addProperty(String name, String type)
    {
        properties.put(name, new Property(name, type));
    }

    public void removeProperty(String name)
    {
        properties.remove(name);
    }

    private String protocolCode;

    public void setProtocolCode(String s)
    {
        protocolCode = s;
    }

    public String getProtocolCode()
    {
        return protocolCode;
    }

    public ArrayType getArrayType(String property)
    {
        ArrayType t = (ArrayType) arrayTypes.get(property);
        if(t == null)
        {
            t = new ArrayType(this, property);
            parent.addChild(t);
            arrayTypes.put(property, t);
        }
        return t;
    }

    void putArrayType(ArrayType arr)
    {
        arrayTypes.put(arr.getProperty(), arr);
    }

    protected void loadCommitedObjects(TIRObject o)
    {
        if(loaded) return;
        loaded = true;
        try
        {
            Container cont = (Container) o;
            Contained objs[] = cont.contents(null);
            for(int i = 0; i < objs.length; i++)
            {
                if(objs[i] instanceof ProvideDef)
                {
                    String arrayFullName = null;
                    CDLType type = ((ProvideDef) objs[i]).provide();
                    if(type instanceof TypedefDef)
                    {
                        arrayFullName = getFullName((TypedefDef) type);
                        type = ((TypedefDef) type).original_type();
                        if(!(type instanceof ArrayDef)) continue;
                        type = ((ArrayDef) type).element_type();
                    }
                    if(type instanceof InterfaceDef)
                    {
                        String name = objs[i].get_identification().name();
                        Interface iface = new Interface(name, getFullName((InterfaceDef) type), arrayFullName, PROVISION);
                        interfacesIndex.put(name, iface);
                        interfaces.add(iface);
                    }
                }
                else if(objs[i] instanceof RequireDef)
                {
                    String arrayFullName = null;
                    CDLType type = ((RequireDef) objs[i]).require();
                    if(type instanceof TypedefDef)
                    {
                        arrayFullName = getFullName((TypedefDef) type);
                        type = ((TypedefDef) type).original_type();
                        if(!(type instanceof ArrayDef)) continue;
                        type = ((ArrayDef) type).element_type();
                    }
                    if(type instanceof InterfaceDef)
                    {
                        String name = objs[i].get_identification().name();
                        Interface iface = new Interface(name, getFullName((InterfaceDef) type), arrayFullName, REQUIREMENT);
                        interfacesIndex.put(name, iface);
                        interfaces.add(iface);
                    }
                }
                else if(objs[i] instanceof PropertyDef)
                {
                    String name = objs[i].get_identification().name();
                    CDLType type = ((PropertyDef) objs[i]).type();
                    if(type instanceof PrimitiveDef)
                    {
                        int kind = ((PrimitiveDef) type).kind().value();
                        if(kind == PrimitiveKind.pk_string || kind == PrimitiveKind.pk_wstring)
                            addProperty(name, "string");
                        else if(kind == PrimitiveKind.pk_long || kind == PrimitiveKind.pk_longlong || kind == PrimitiveKind.pk_short
                                || kind == PrimitiveKind.pk_ulong || kind == PrimitiveKind.pk_ulonglong || kind == PrimitiveKind.pk_ushort)
                            addProperty(name, "integer");
                    }
                }
            }
        }
        catch (Exception e)
        {
            System.out.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    private Collection interfaces;
    private Map interfacesIndex;
    private Map arrayTypes;
    private Map properties;
    protected static int counter;
    private boolean loaded = false;


    private String getFullName(Contained obj) throws RemoteException
    {
        Identification id = obj.get_identification();
        String absoluteName = id.absolute_name().name();
        return absoluteName.substring(absoluteName.indexOf("::"))
                + " ver. " + id.version();
    }

    String writeAsCDL(int indent)
    {

        if(commited) return "";
        if(getName().equalsIgnoreCase("SYSTEM")) return "";

        String s = "\n" + indentString(indent) + "frame " + getName() + " {\n";

        // write properties
        Property[] prop = getProperties();
        for(int i = 0; i < prop.length; i++)
            s += indentString(indent + 1) + "property " + prop[i].getType()
                    + " " + prop[i].getName() + ";\n";

        // write provisions
        Interface[] pro = getProvisions();
        if(pro.length > 0)
            s += indentString(indent + 1) + "provides:\n";
        for(int i = 0; i < pro.length; i++)
            if(pro[i].getArray() != null)
                s += indentString(indent + 2)
                        + pro[i].getArray().getRelativeName(getParent())
                        + " " + pro[i].getName() + ";\n";
            else
                s += indentString(indent + 2)
                        + pro[i].getType().getRelativeName(getParent())
                        + " " + pro[i].getName() + ";\n";

        // write requirements
        Interface[] req = getRequirements();
        if(req.length > 0)
            s += indentString(indent + 1) + "requires:\n";
        for(int i = 0; i < req.length; i++)
            if(req[i].getArray() != null)
                s += indentString(indent + 2)
                        + req[i].getArray().getRelativeName(getParent())
                        + " " + req[i].getName() + ";\n";
            else
                s += indentString(indent + 2)
                        + req[i].getType().getRelativeName(getParent())
                        + " " + req[i].getName() + ";\n";

        // write protocol
        if(protocolCode != null && !protocolCode.equals(""))
        {
            s += "\n" + indentString(indent + 1) + "protocol:\n";
            s += indentCode(protocolCode, indent + 2);
        }

        s += indentString(indent) + "};\n";

        ArrayType[] arr = (ArrayType[]) arrayTypes.values().toArray(new ArrayType[0]);
        for(int i = 0; i < arr.length; i++)
            s += "\n" + arr[i].writeAsCDL(indent);

        return s;
    }
}